// 
// AYA version 5
//
// 1DLLNX@CLib1
// written by umeici. 2004
// 

#ifndef POSIX
#  include "stdafx.h"
#else
#  include "posix_utils.h"
#  include <dlfcn.h>
#  include <stdlib.h>
#  include <sys/types.h>
#  include <sys/stat.h>
#  include <vector>
#endif
#include "lib.h"
#include "misc.h"
#include "ccct.h"
#include "log.h"

#ifndef POSIX
bool	(*loadlib)(HGLOBAL h, long len);
bool	(*unloadlib)(void);
HGLOBAL	(*requestlib)(HGLOBAL h, long *len);
#else
int     (*loadlib)(char* h, long len);
int     (*unloadlib)(void);
char*   (*requestlib)(char* h, long *len);
#endif

extern CLog logger;

#ifdef POSIX
static string str_getenv(const string& name) {
    char* var = getenv(name.c_str());
    if (var == NULL) {
        return string();
    }
    else {
        return string(var);
    }
}
// dllT[`pX֘AB
// POSIXł́ȀꏊDLLƓ̃CuuSAORIɑΉB
// DLLƓƂ͉]ĂA̓V{bNNłׂŁAႦΎ̂悤ɂłB
// % pwd
// /home/foo/.saori
// % ls -l
// -rwxr-xr-x x foo bar xxxxx 1 1 00:00 libssu.so
// lrwxr-xr-x x foo bar xxxxx 1 1 00:00 ssu.dll -> libssu.so
//
// pX͊ϐ SAORI_FALLBACK_PATH 擾B̓R؂̐΃pXłB
static vector<string> posix_dll_search_path;
static bool posix_dll_search_path_is_ready = false;
static string posix_search_fallback_dll(const string& dllfile) {
    // dllfile͒Tt@CDLLBpX؂蕶/B
    // փCut΂̐΃pXA
    // tȂ΋󕶎ԂB
    
    if (!posix_dll_search_path_is_ready) {
	// SAORI_FALLBACK_PATHB
	string path = str_getenv("SAORI_FALLBACK_PATH");
        if (path.length() > 0) {
            while (true) {
                string::size_type colon_pos = path.find(':');
                if (colon_pos == string::npos) {
                    posix_dll_search_path.push_back(path);
                    break;
                }
                else {
                    posix_dll_search_path.push_back(path.substr(0, colon_pos));
                    path.erase(0, colon_pos+1);
                }
            }
        }
        posix_dll_search_path_is_ready = true;
    }

    string::size_type pos_slash = dllfile.rfind('/');
    string fname(
	dllfile.begin() + (pos_slash == string::npos ? 0 : pos_slash),
	dllfile.end());

    for (vector<string>::const_iterator ite = posix_dll_search_path.begin();
	 ite != posix_dll_search_path.end(); ite++ ) {
	string fpath = *ite + '/' + fname;
	struct stat sb;
	if (stat(fpath.c_str(), &sb) == 0) {
	    // փCu݂悤Bȏ̃`FbN͏ȗB
	    return fpath;
	}
    }
    return string();
}
#endif

/* -----------------------------------------------------------------------
 *  ֐  F  CLib1::LoadLib
 *  @\TvF  DLL[h܂
 *
 *  Ԓl@@F@0/1=s/(Ƀ[hĂ܂)
 * -----------------------------------------------------------------------
 */
#ifndef POSIX
int	CLib1::LoadLib(void)
{
	if (hDLL != NULL)
		return 1;

	char	*dllpathname = Ccct::Ucs2ToMbcs(name, CHARSET_DEFAULT);
	if (dllpathname == NULL)
		return 0;

	hDLL = LoadLibrary(dllpathname);
	free(dllpathname);
	
	return (hDLL != NULL) ? 1 : 0;
}
#else
int CLib1::LoadLib() {
    if (hDLL != NULL) {
	return 1;
    }

    string libfile = narrow(name);
    fix_filepath(libfile);

    // ϐ SAORI_FALLBACK_ALWAYS `ĂāA
    // ł"0"łȂ΁Adllt@CJĂ݂鎖
    // ߂ȂBłȂ΁AdlopenĂ݂B
    char* env_fallback_always = getenv("SAORI_FALLBACK_ALWAYS");
    bool fallback_always = false;
    if (env_fallback_always != NULL) {
	string str_fallback_always(env_fallback_always);
	if (str_fallback_always.length() > 0 &&
	    str_fallback_always != "0") {
	    fallback_always = true;
	}
    }
    bool do_fallback = true;
    if (!fallback_always) {
	void* handle = dlopen(libfile.c_str(), RTLD_LAZY);
	if (handle != NULL) {
	    // load, unload, requestoĂ݂B
	    void* sym_load = dlsym(handle, "load");
	    void* sym_unload = dlsym(handle, "unload");
	    void* sym_request = dlsym(handle, "request");
	    if (sym_load != NULL && sym_unload != NULL && sym_request != NULL) {
		// ȂƐɓǂ߂B
		do_fallback = false;
	    }
	}
	dlclose(handle);
    }
    if (do_fallback) {
	// փCuTB
	string fallback_lib = posix_search_fallback_dll(get_fname(libfile));
	if (fallback_lib.length() == 0) {
	    // B
	    char* cstr_path = getenv("SAORI_FALLBACK_PATH");
	    string fallback_path =
		(cstr_path == NULL ?
		 "(environment variable `SAORI_FALLBACK_PATH' is empty)" : cstr_path);
	    string message =
		libfile+": This is not usable in this platform.\n"+
		"Fallback library doesn't exist: "+fallback_path+"\n";
	    logger.Write(widen(message));
	    return 0;
	}
	else {
	    string message =
		"SAORI: using "+fallback_lib+" instead of "+libfile+"\n";
	    logger.Write(widen(message));
	    
	    libfile = fallback_lib;
	}
    }

    hDLL = dlopen(libfile.c_str(), RTLD_LAZY);
    return (hDLL != NULL) ? 1 : 0;
}
#endif

/* -----------------------------------------------------------------------
 *  ֐  F  CLib1::Load
 *  @\TvF  loads܂
 *
 *  Ԓl@@F@0/1=s/
 * -----------------------------------------------------------------------
 */
#ifndef POSIX
int	CLib1::Load(void)
{
	if (hDLL == NULL)
		return 0;

	// AhX擾
	loadlib = (bool (*)(HGLOBAL h, long len))GetProcAddress(hDLL, "load");
	if (loadlib == NULL)
		return 0;

	// DLLpX쐬
	wchar_t	drive[_MAX_DRIVE], dir[_MAX_DIR], fname[_MAX_FNAME], ext[_MAX_EXT];
	_wsplitpath(name.c_str(), drive, dir, fname, ext);
	wstring	dllpath = drive;
	dllpath += dir;

	// pXMBCSɕϊ
	char	*t_dllpath = Ccct::Ucs2ToMbcs(dllpath, charset);
	if (t_dllpath == NULL)
		return 0;

	long	len = (long)strlen(t_dllpath);

	// pXq[vɃRs[
	HGLOBAL	gmem = ::GlobalAlloc(GMEM_FIXED, len);
	memcpy(gmem, t_dllpath, len);
	free(t_dllpath);

	// s
	(*loadlib)(gmem, len);
	
	return 1;
}
#else
int CLib1::Load(void) {
    if (hDLL == NULL) {
	return 0;
    }
    
    // AhX擾
    loadlib = reinterpret_cast<int(*)(char*,long)>(dlsym(hDLL, "load"));
    if (loadlib == NULL) {
	return 0;
    }
    
    // DLLpX쐬
    wstring::size_type pos_slash = name.rfind(L'/');
    string dllpath;
    if (pos_slash == wstring::npos) {
	dllpath = ".";
    }
    else {
	dllpath = narrow(name.substr(0, pos_slash+1));
    }

    long len = dllpath.length();

    // pXq[vɃRs[
    char* gmem = static_cast<char*>(malloc(len));
    memcpy(gmem, dllpath.c_str(), len);

    // s
    (*loadlib)(gmem, len);
    
    return 1;
}
#endif

/* -----------------------------------------------------------------------
 *  ֐  F  CLib1::Unload
 *  @\TvF  unloads܂
 *
 *  Ԓl@@F@0/1/2=s//[hĂȂA͊unloadĂ
 * -----------------------------------------------------------------------
 */
#ifndef POSIX
int	CLib1::Unload(void)
{
	if (hDLL == NULL)
		return 2;

	// AhX擾
	unloadlib = (bool (*)(void))GetProcAddress(hDLL, "unload");
	if (unloadlib == NULL)
		return 0;

	// s
	(*unloadlib)();

	return 1;
}
#else
int CLib1::Unload(void) {
    if (hDLL == NULL) {
	return 2;
    }

    // AhX擾
    unloadlib = reinterpret_cast<int(*)(void)>(dlsym(hDLL, "unload"));
    if (unloadlib == NULL) {
	return 0;
    }

    // s
    (*unloadlib)();
    
    return 1;
}
#endif

/* -----------------------------------------------------------------------
 *  ֐  F  CLib1::Release
 *  @\TvF  DLL[X܂
 * -----------------------------------------------------------------------
 */
#ifndef POSIX
void	CLib1::Release(void)
{
	if (hDLL == NULL)
		return;

	FreeLibrary(hDLL);
	hDLL = NULL;
}
#else
void CLib1::Release(void) {
    if (hDLL == NULL) {
	return;
    }

    dlclose(hDLL);
    hDLL = NULL;
}
#endif

/* -----------------------------------------------------------------------
 *  ֐  F  CLib1::Request
 *  @\TvF  requests܂
 *
 *  Ԓl@@F@0/1=s/
 * -----------------------------------------------------------------------
 */
#ifndef POSIX
int	CLib1::Request(const wstring &istr, wstring &ostr)
{
	ostr = L"";

	if (hDLL == NULL)
		return 0;

	// AhX擾
	requestlib = (HGLOBAL (*)(HGLOBAL h, long *len))GetProcAddress(hDLL, "request");
	if (requestlib == NULL)
		return 0;

	// }`oCgR[hɕϊ
	char	*t_istr = Ccct::Ucs2ToMbcs(istr, charset);
	if (t_istr == NULL)
		return 0;

	long	len = (long)strlen(t_istr);

	// pXq[vɃRs[
	HGLOBAL	igmem = ::GlobalAlloc(GMEM_FIXED, len);
	memcpy(igmem, t_istr, len);
	free(t_istr);

	// s
	HGLOBAL	ogmem = (*requestlib)(igmem, &len);

	// ʎ擾
	char	*t_ostr = (char *)malloc((len + 1)*sizeof(char));
	if (t_ostr == NULL) {
		GlobalFree(ogmem);
		return 0;
	}
	strncpy(t_ostr, (char *)ogmem, len);
	*(t_ostr + len) = '\0';
	GlobalFree(ogmem);

	// ʂUCS-2֕ϊ
	wchar_t	*t_ostr2 = Ccct::MbcsToUcs2(t_ostr, charset);
	free(t_ostr);
	if (t_ostr2 == NULL)
		return 0;
	ostr = t_ostr2;
	free(t_ostr2);

	return 1;
}
#else
int CLib1::Request(const wstring &istr, wstring &ostr) {
    ostr = L"";

    if (hDLL == NULL) {
	return 0;
    }
    
    // AhX擾
    requestlib = reinterpret_cast<char*(*)(char*, long *)>(dlsym(hDLL, "request"));
    if (requestlib == NULL) {
	return 0;
    }
    
    // }`oCgR[hɕϊ
    char *t_istr = Ccct::Ucs2ToMbcs(istr, charset);
    if (t_istr == NULL) {
	return 0;
    }

    long len = (long)strlen(t_istr);

    // pXq[vɃRs[
    char* igmem = static_cast<char*>(malloc(len));
    memcpy(igmem, t_istr, len);
    free(t_istr);

    // s
    char* ogmem = (*requestlib)(igmem, &len);

    // ʎ擾
    string t_ostr(ogmem, len);

    // ʂUCS-2֕ϊ
    wchar_t *t_ostr2 = Ccct::MbcsToUcs2(t_ostr, charset);
    if (t_ostr2 == NULL) {
	return 0;
    }
    ostr = t_ostr2;
    free(t_ostr2);

    return 1;
}
#endif
